package be.iminds.ilabt.jfed.highlevel.jobs;

import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.ExperimentPart;
import be.iminds.ilabt.jfed.experiment.SfaExperimentPart;
import be.iminds.ilabt.jfed.experiment.tasks.ExperimentTaskStatus;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.highlevel.DesiredStatus;
import be.iminds.ilabt.jfed.highlevel.controller.TaskThread;
import be.iminds.ilabt.jfed.highlevel.jobs.parts.ExperimentPartController;
import be.iminds.ilabt.jfed.highlevel.jobs.parts.ExperimentPartControllerManager;
import be.iminds.ilabt.jfed.highlevel.jobs.parts.NonSfaExperimentPartController;
import be.iminds.ilabt.jfed.highlevel.jobs.states.*;
import be.iminds.ilabt.jfed.highlevel.model.InternalState;
import be.iminds.ilabt.jfed.highlevel.tasks.HighLevelTaskFactory;
import be.iminds.ilabt.jfed.highlevel.util.AggregateManagerWrapperFactory;
import be.iminds.ilabt.jfed.highlevel.util.ProxySocketFactoryProvider;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.AggregateManagerWrapper;
import be.iminds.ilabt.jfed.rspec.model.*;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXModelRspec;
import be.iminds.ilabt.jfed.rspec.rspec_source.ManifestRspecSource;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import be.iminds.ilabt.jfed.util.BestNodeLoginFinder;
import be.iminds.ilabt.jfed.util.SshProxySocketImplFactory;
import net.schmizz.sshj.SSHClient;
import net.schmizz.sshj.transport.TransportException;
import net.schmizz.sshj.transport.verification.PromiscuousVerifier;
import net.schmizz.sshj.userauth.UserAuthException;
import net.schmizz.sshj.userauth.keyprovider.KeyPairWrapper;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.net.SocketFactory;
import java.io.IOException;
import java.security.KeyPair;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;

/**
 * This executes all actions that jFed needs to do AFTER an experiment is READY:
 *
 * - Sharing LANs (if specified in RSpec)
 * - Waiting for startup scripts (if wait method specified in RSpec)
 * - Distributing SSH keys (if specified in RSpec)
 * - Executing ansible (if ansible stuff specified in RSpec)
 * - Execute the dir, upload and execute steps of the ExperimentSpecification (if present)
 *
 * This does not:
 * - Test Connectivity  (should be done BEFORE this step)
 */
public class SetupSoftwareExperimentJob extends AbstractJob<Boolean> {
    private static final Logger LOG = LoggerFactory.getLogger(SetupSoftwareExperimentJob.class);

    @Nonnull private final JobStateFactory jobStateFactory;

    @Nonnull private final AggregateManagerWrapperFactory aggregateManagerWrapperFactory;
    @Nonnull private final ExperimentPartControllerManager epcManager;
    @Nonnull private final ProxySocketFactoryProvider proxySocketFactoryProvider;
    @Nonnull private final GeniUserProvider geniUserProvider;

    private Boolean success = null;

    public SetupSoftwareExperimentJob(@Nonnull Experiment experiment,
                                      @Nonnull HighLevelTaskFactory hltf,
                                      @Nonnull TaskThread tt,
                                      @Nonnull JobStateFactory jobStateFactory,
                                      @Nonnull AggregateManagerWrapperFactory aggregateManagerWrapperFactory,
                                      @Nonnull ExperimentPartControllerManager epcManager,
                                      @Nonnull ProxySocketFactoryProvider proxySocketFactoryProvider,
                                      @Nonnull GeniUserProvider geniUserProvider) {
        super("Setup Software for Experiment " + experiment.getName(), experiment, hltf, tt);
        this.jobStateFactory = jobStateFactory;
        this.aggregateManagerWrapperFactory = aggregateManagerWrapperFactory;
        this.epcManager = epcManager;
        this.proxySocketFactoryProvider = proxySocketFactoryProvider;
        this.geniUserProvider = geniUserProvider;

        assert experiment.getRequestedEndTime() != null;
    }

    @Override
    public Boolean execute() throws Exception {
        if (!hasSfaParts()) {
            if (experiment.getExperimentSpecification() != null) {
                experiment.getExperimentSpecificationLogger().firePostRSpecFail(experiment.getExperimentSpecification().getRspecs().get(0),
                        "assert failed: experiment.getExperimentSpecification() == null", null);
                experiment.getExperimentSpecificationLogger().firePostRspecAll(false, Collections.emptyList());
            }
            return true;
        }

        LOG.debug("SetupSoftwareExperimentJob started");

        //Assumption: everything is now successfully running and we have connectivity available

        if (experiment.getSliceOrNull() == null ||
                experiment.getSlice().getRequestRspec() == null ||
                experiment.getSlice().getRequestRspec().getModelRspec() == null) {
            LOG.warn("Problem with experiment.getSlice().getRequestRspec(). " +
                    "Will assume this is normal and ABORT software setup.");

            if (experiment.getExperimentSpecification() != null) {
                experiment.getExperimentSpecificationLogger().firePostRSpecSuccess(
                        experiment.getExperimentSpecification().getRspecs().get(0));
                experiment.getExperimentSpecificationLogger().firePostRspecAll(
                        true, getExperimentClientIds());
            }

            success = true;
            return true;
        }

        boolean hasAutoShareLans = false;
        Collection<? extends RspecLink> links = experiment.getSlice()
                .getRequestRspec().getModelRspec().getLinks();
        for (RspecLink link : links) {
            List<String> autoShareLanNames = link.getAutoShareLanNames();
            //noinspection ConstantConditions
            if (autoShareLanNames != null && !autoShareLanNames.isEmpty()) {
                hasAutoShareLans = true;
                break;
            }
        }
        if (hasAutoShareLans) {
            LOG.info("Experiment has auto shared LANs. Will share them.");
            final AutoShareLanPartsState autoShareLanPartsState = new AutoShareLanPartsState();
            setAndRunState(autoShareLanPartsState);

            if (autoShareLanPartsState.getStatus() != ExperimentTaskStatus.SUCCESS) {
                LOG.info("Failure sharing LANs");

                if (experiment.getExperimentSpecification() != null) {
                    experiment.getExperimentSpecificationLogger().firePostRSpecFail(
                            experiment.getExperimentSpecification().getRspecs().get(0),
                            "Failure sharing LANs", null);
                    experiment.getExperimentSpecificationLogger().firePostRspecAll(
                            false, getExperimentClientIds());
                }

                success = false;
                return false;
            }

            LOG.debug("Experiment auto shared LANs have been shared. Will wait for ready again.");
            final WaitForOpStatusExperimentJob.UpdateSfaPartsUntilTargetOpStatus updateUntilReadyState2 =
                    new WaitForOpStatusExperimentJob.UpdateSfaPartsUntilTargetOpStatus(
                            jobStateFactory, this,
                            DesiredStatus.GENI_READY);
            setAndRunState(updateUntilReadyState2);

            if (updateUntilReadyState2.isPermanentFailure()) {
                LOG.info("WaitForPartsReadyState (after auto shared LANs) reported permanent failure");

                if (experiment.getExperimentSpecification() != null) {
                    experiment.getExperimentSpecificationLogger().firePostRSpecFail(
                            experiment.getExperimentSpecification().getRspecs().get(0),
                            "WaitForPartsReadyState (after auto shared LANs) reported permanent failure", null);
                    experiment.getExperimentSpecificationLogger().firePostRspecAll(
                            false, getExperimentClientIds());
                }

                success = false;
                return false;
            }

            LOG.debug("All is ready after sharing LANs");
        } else
            LOG.debug("Experiment has NO auto shared LANs");


        if (!experiment.getKeypairs().isEmpty() && DistributeSshKeysState.hasDistributeSshKeyPair(experiment)) {
            LOG.debug("Experiment requires a distributed keypair. Will generate and distribute it.");
            setAndRunSshUsingState(jobStateFactory.createDistributeSshKeysState(this));
        } else {
            LOG.debug("Experiment does not require a distributed keypair.");
        }

        if (WaitForExecuteFinishedState.hasWaitForExecute(experiment)) {
            LOG.debug("Experiment requires waiting for some server executes to finish.");
            setAndRunSshUsingState(jobStateFactory.createWaitForExecuteFinishedState(this));
        } else {
            LOG.debug("Experiment does not require waiting for server execute.");
        }

        if (ExecuteAnsibleServicesState.hasAnsibleServices(experiment)) {
            LOG.debug("Experiment requires executing ansible scripts.");
            setAndRunSshUsingState(jobStateFactory.createExecuteAnsibleServicesState(this));
        } else {
            LOG.debug("Experiment does not require executing ansible scripts.");
        }

        if (experiment.getExperimentSpecification() != null) {
            LOG.info("Experiment run with ExperimentSpecification. Will now start its dir, upload and execute phases.");

            experiment.getExperimentSpecificationLogger().firePostRSpecSuccess(
                    experiment.getExperimentSpecification().getRspecs().get(0));
            experiment.getExperimentSpecificationLogger().firePostRspecAll(
                    true, getExperimentClientIds());

            List<State> states = jobStateFactory.createOrderedExperimentServicesStates(this);
            for (State state: states) {
                setAndRunSshUsingState(state);
                if (state.getStatus() != ExperimentTaskStatus.SUCCESS) {
                    //INACTIVE, BUSY, FAILED, WARNING
                    experiment.getExperimentSpecificationLogger().firePostExecuteAll(false);
                    success = false;
                    return false;
                }
            }
            experiment.getExperimentSpecificationLogger().firePostExecuteAll(true);
        } else {
            LOG.info("Experiment run without ExperimentSpecification.");
        }

        LOG.debug("SetupSoftwareExperimentJob stopped successfully");

        success = true;
        return true;
    }

    protected void setAndRunSshUsingState(@Nonnull State newState) throws JFedException, InterruptedException {
        boolean sawException = true;
        try {
            sawException = true;
            super.setAndRunState(newState);
            sawException = false;
        } catch (Throwable t) {
            sawException = true;
            throw t;
        } finally {
            if (sawException || newState.getStatus() != ExperimentTaskStatus.SUCCESS) {
                success = false;
                LOG.warn("Something went wrong in "+newState.getName()+" ("+newState.getStatus()+") (ex="+sawException+") -> closing all SSH connections");
                closeAllSshConnections();
                //a bit dirty, but this works (and shouldn't mask exceptions)
                if (!sawException) {
                    throw new RuntimeException("SetupSoftwareExperimentJob encountered a failure in: " + newState.getName());
                }
            }
        }
    }

    @Nonnull
    private List<String> getExperimentClientIds() {
        if (experiment.getSliceOrNull() == null || experiment.getSliceOrNull().getRequestRspec() == null) {
            return Collections.emptyList();
        }
        RequestRspecSource requestRspecSource = experiment.getSliceOrNull().getRequestRspec();
        FXModelRspec model = requestRspecSource == null ?
                null :
                requestRspecSource.getModelRspec(ModelRspecType.FX);
        assert model != null;

        if (model == null) {
            return Collections.emptyList();
        }

        return model.getNodes().stream().map(RspecNode::getClientId).filter(Objects::nonNull).collect(Collectors.toList());
    }

    public Boolean isSuccess() {
        return success;
    }

    public class AutoShareLanPartsState extends SlicedState<AutoShareLanPartStateSlice> {
        private final List<AutoShareLanPartStateSlice> slices;

        protected AutoShareLanPartsState() {
            super("Auto Sharing LANs in all ExperimentParts");

            slices = new ArrayList<>();

            List<SfaExperimentPart> sfaExperimentParts =
                    experiment.getPartsStream()
                            .filter(part -> part.getState() != InternalState.FAILED)
                            .filter(part -> part instanceof SfaExperimentPart)
                            .map(part -> (SfaExperimentPart) part)
                            .collect(Collectors.toList());

            experiment.requireSlice();
            if (experiment.getSlice().getRequestRspec() == null)
                throw new IllegalStateException("experiment.getSlice().getRequestRspec() is null which is not supposed to occur in this state");
            if (experiment.getSlice().getRequestRspec().getModelRspec() == null)
                throw new IllegalStateException("experiment.getSlice().getRequestRspec().getModelRspec() is null which is not supposed to occur in this state");
            if (experiment.getSlice().getRequestRspec().getModelRspec().getLinks() == null)
                throw new IllegalStateException("experiment.getSlice().getRequestRspec().getModelRspec().getLinks() is null which is not supposed to occur in this state");

            Collection<? extends RspecLink> links = experiment.getSlice().getRequestRspec().getModelRspec().getLinks();

            for (SfaExperimentPart sfaExperimentPart : sfaExperimentParts) {
                Server auth = sfaExperimentPart.getConnectSfaAuthority();
                boolean hasAutoShareLans = false;

                for (RspecLink link : links) {
                    List<String> autoShareLanNames = link.getAutoShareLanNames();
                    //noinspection ConstantConditions
                    if (autoShareLanNames != null && !autoShareLanNames.isEmpty() && link.getComponentManagerUrns().contains(auth.getDefaultComponentManagerAsGeniUrn())) {
                        hasAutoShareLans = true;
                        break;
                    }
                }

                if (hasAutoShareLans) {
                    LOG.debug("Auto Sharing LAN(s) in " + auth.getName());
                    AutoShareLanPartStateSlice slice = new AutoShareLanPartStateSlice(sfaExperimentPart);
                    slices.add(slice);
                } else {
                    LOG.debug("No Auto Shared LAN  in " + auth.getName());
                }
            }

            LOG.debug("AutoShareLanPartsState constructor finished with " + slices.size() + " slices.");

            assert !slices.isEmpty();
        }

        @Override
        public Collection<AutoShareLanPartStateSlice> getSlices() {
            return slices;
        }
    }

    public class AutoShareLanPartStateSlice extends ExperimentPartStateSlice<SfaExperimentPart> {

        public AutoShareLanPartStateSlice(SfaExperimentPart experimentPart) {
            super(SetupSoftwareExperimentJob.this, experimentPart);
        }

        @Override
        public ExperimentTaskStatus statefulRun() throws JFedException, InterruptedException {
            if (!supportsShareLan(experimentPart.getConnectSfaAuthority())) {
                LOG.warn("Share LAN on {} is not supported. Will try anyway.", experimentPart.getName());
                //return ExperimentTaskStatus.WARNING;
            }
            experiment.requireSlice();
            if (experiment.getSlice().getRequestRspec() == null)
                throw new IllegalStateException("experiment.getSlice().getRequestRspec() is null which is not supposed to occur in this state");
            if (experiment.getSlice().getRequestRspec().getModelRspec() == null)
                throw new IllegalStateException("experiment.getSlice().getRequestRspec().getModelRspec() is null which is not supposed to occur in this state");
            if (experiment.getSlice().getRequestRspec().getModelRspec().getLinks() == null)
                throw new IllegalStateException("experiment.getSlice().getRequestRspec().getModelRspec().getLinks() is null which is not supposed to occur in this state");

            Collection<? extends RspecLink> links = experiment.getSlice().getRequestRspec().getModelRspec().getLinks();

            ExperimentTaskStatus worstStatus = null;
            Server auth = experimentPart.getConnectSfaAuthority();
            LOG.debug("Start sharing LAN(s) on " + auth.getDefaultComponentManagerAsGeniUrn());
            for (RspecLink link : links) {
                List<String> autoShareLanNames = link.getAutoShareLanNames();
                //noinspection ConstantConditions
                if (autoShareLanNames != null && !autoShareLanNames.isEmpty() && link.getComponentManagerUrns().contains(auth.getDefaultComponentManagerAsGeniUrn())) {
                    String sharedLanName = autoShareLanNames.get(0);

                    if (autoShareLanNames.size() > 1)
                        LOG.warn("There is more than one auto share LAN name per link (for link\"" + link.getClientId() + "\"), ignoring all but first.");

                    LOG.debug("Sharing LAN " + link.getClientId() + " as " + sharedLanName + " on " + auth.getDefaultComponentManagerAsGeniUrn());

                    ShareLanAtAuthorityPartState shareLanState =
                            jobStateFactory.createShareLanState(SetupSoftwareExperimentJob.this, experimentPart, link.getClientId(), sharedLanName);

                    setAndRunState(shareLanState);
                    ExperimentTaskStatus status = shareLanState.getStatus();
                    LOG.debug("Shared LAN " + link.getClientId() + " as " + sharedLanName + " on " + auth.getDefaultComponentManagerUrn() + " -> status = " + status);

                    if (status != ExperimentTaskStatus.SUCCESS || worstStatus == null) {
                        worstStatus = status;
                    }
                }
            }

            //there should be at least one link shared, or this shouldn't have been called.
            assert worstStatus != null;

            LOG.debug("Stopped sharing LAN(s) on " + auth.getDefaultComponentManagerUrn() + " -> worst status = " + worstStatus);

            return worstStatus;
        }

        private boolean supportsShareLan(Server auth) {
            AggregateManagerWrapper amWrapper =
                    aggregateManagerWrapperFactory.getAggregateManagerWrapper(auth);

            return amWrapper.isShareLanSupported();
        }
    }





    private KeyPair loginKeyPair;
    private String loginUsername;

    @Nonnull
    public KeyPair getKeyPair() {
        if (this.loginKeyPair == null) {
            //find the keypair and username to login. This is either the distributed key, or if that is not present, just the user's key.
            assert experiment.getKeypairs() != null;
            loginUsername = null;
            KeyPair loginKeyPair = null;

            for (Map.Entry<DistributeSshKeypair, KeyPair> e : experiment.getKeypairs().entrySet()) {
                if(e.getValue() != null){
                    //distributedSshKeysUsername may be null, but that's ok, that just means the logged in users account needs to be used.
                    loginUsername = e.getKey().getUser() == null ? null : e.getKey().getUser().getResourceName();

                    loginKeyPair = e.getValue();
                    break;
                }
            }

            if(loginKeyPair == null) {
                loginKeyPair = new KeyPair(geniUserProvider.getLoggedInGeniUser().getPublicKey(),
                        geniUserProvider.getLoggedInGeniUser().getPrivateKey());
                loginUsername = geniUserProvider.getLoggedInGeniUser().getUserUrn().getEncodedResourceName();
            }

            this.loginKeyPair = loginKeyPair;
        }

        return this.loginKeyPair;
    }

    @Nonnull
    public String getLoginUsername() {
        if (this.loginUsername == null || this.loginKeyPair == null) {
            getKeyPair();
            if (loginUsername == null)
                throw new IllegalStateException("Did not find login username");
        }
        return loginUsername;
    }

    private BestNodeLoginFinder bestNodeLoginFinder;
    @Nullable
    public BestNodeLoginFinder getBestNodeLoginFinder() {
        if (bestNodeLoginFinder == null) {
            String distributedSshKeysUsername = getLoginUsername();

            ManifestRspecSource manifestRspecSource = experiment.getSlice().getManifestRspec();
            bestNodeLoginFinder = manifestRspecSource == null ||
                    manifestRspecSource.getStringRspec() == null ? null :
                    new BestNodeLoginFinder(manifestRspecSource.getStringRspec(),
                            distributedSshKeysUsername,
                            this.geniUserProvider.getLoggedInGeniUser(),
                            new BestNodeLoginFinder.Feedback() {
                                @Override
                                public void info(String s) {
                                    LOG.info("SetupSoftwareExperimentJob.BestNodeLoginFinder: " + s);
                                }

                                @Override
                                public void error(String s) {
                                    LOG.error("SetupSoftwareExperimentJob.BestNodeLoginFinder: " + s);
                                }

                                @Override
                                public void debug(String s) {
                                    LOG.debug("SetupSoftwareExperimentJob.BestNodeLoginFinder: " + s);
                                }
                            });
        }

        return bestNodeLoginFinder;
    }

    public static class SshException extends Exception {
        public SshException() { }
        public SshException(String s) { super(s); }
        public SshException(String s, Throwable throwable) { super(s, throwable); }
        public SshException(Throwable throwable) { super(throwable); }
        public SshException(String s, Throwable throwable, boolean b, boolean b1) {
            super(s, throwable, b, b1);
        }
    }

    private Map<String, SSHClient> sshClientByUniqueNodeId = new ConcurrentHashMap<>();

    /**
     * Will return a connected SSH connection. This can return an existing connection,
     * and will create a new one if needed.
     *
     * @param node node to which a connection is required
     * @param stateSlice may be null. If not, setAndRunState might be run on it to setup the connection
     * @return a connected SSHConnection. (never returns null)
     * @throws JFedException
     * @throws InterruptedException
     * @throws SshException if anything went wrong seting up the connection, or is wrong with the existing connection
     */
    @Nonnull
    public synchronized SSHClient getSSHClient(@Nonnull RspecNode node, @Nullable StateSlice stateSlice)
            throws JFedException, InterruptedException, SshException {
        SSHClient res = sshClientByUniqueNodeId.get(node.getUniqueId());
        if (res != null) {
            if (!res.isConnected()) {
                throw new SshException("SSH connection is not connected");
            }

            return res;
        }

        final SetupSshConnectionState setupSshConnectionState = new SetupSshConnectionState(node);
        if (stateSlice == null) {
            setupSshConnectionState.executeState(this);
        } else {
            stateSlice.setAndRunState(setupSshConnectionState);
            if (setupSshConnectionState.getStatus() != ExperimentTaskStatus.SUCCESS) {
                throw new SshException("Failed to setup SSH connection ("+setupSshConnectionState.getStatus()+")");
            }
        }
        res = setupSshConnectionState.getSsh();
        if (res == null) {
            throw new SshException("Failed to setup SSH connection (sshClient=null)");
        }
        sshClientByUniqueNodeId.put(node.getUniqueId(), res);

        if (!res.isConnected()) {
            throw new SshException("SSH connection is not connected");
        }
        return res;
    }

    /**
     * Possibly close the SSH connection. Call this if you don't need a connection anymore.
     *
     * @param node
     */
    public synchronized void closeSSH(@Nonnull RspecNode node) {
        //TODO: allow close only if needed

//        SSHClient ssh = sshClientByUniqueNodeId.remove(node.getUniqueId());
//        if (ssh == null)
//            return;
//
//        try {
//            ssh.close();
//        } catch (IOException ignored) {
//        }
    }

    public synchronized void closeAllSshConnections() {
        LOG.debug("closeAllSshConnections() called");
        synchronized (sshClientByUniqueNodeId) {
            for (SSHClient ssh : sshClientByUniqueNodeId.values()) {
                try {
                    if (ssh != null)
                        ssh.close();
                } catch (IOException ignored) {
                    //probably, connection is already closed. We don't really need to worry about this.
                    LOG.debug("Will ignore error while closing SSH connection: " + ignored.getMessage());
                }
            }
            //make sure it isn't reused
            sshClientByUniqueNodeId.clear();
        }
    }


    ////////////////////////////////////////////////////////

    private static final int SETUP_SSH_CONNECTION_RETRIES = 3;

    public class SetupSshConnectionState extends State {

        @Nonnull private final RspecNode node;
        private SSHClient ssh;

        protected SetupSshConnectionState(@Nonnull RspecNode node) {
            super("Setting up SSH Connection to '" + node.getClientId() + "'");
            this.node = node;
        }

        @Nonnull
        @Override
        protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
            try {
                ssh = setupNewSSHClient(node);
            } catch (SshSetupException e) {
                return ExperimentTaskStatus.FAILED;
            }

            return ExperimentTaskStatus.SUCCESS;

        }

        public SSHClient getSsh() {
            return ssh;
        }
    }

    ////////////////////////////////////////////////////////


    public static class SshSetupException extends Exception {
        public SshSetupException() { }
        public SshSetupException(String message) { super(message); }
        public SshSetupException(String message, Throwable cause) { super(message, cause); }
        public SshSetupException(Throwable cause) { super(cause); }
        public SshSetupException(String message, Throwable cause, boolean enableSuppression, boolean writableStackTrace) {
            super(message, cause, enableSuppression, writableStackTrace);
        }
    }

    @Nonnull
    public SSHClient setupNewSSHClient(@Nonnull RspecNode node) throws SshSetupException {
        BestNodeLoginFinder loginFinder = getBestNodeLoginFinder();
        if (loginFinder == null) {
            LOG.warn("loginFinder is null! Will not be able to setup SSH connection.");
        }

        //get login-service associated to the user we are distributing the key to
        BasicStringRspec.LoginService loginService =
                loginFinder == null ?
                        null :
                        loginFinder.findBestLogin(node.getUniqueId());

        if (loginService == null) {
            LOG.warn("Cannot setup connection to {} as no login-service is available", node.getClientId());
            throw new SshSetupException("Cannot setup connection to "+node.getClientId()+" as no login-service is available");
//            return ExperimentTaskStatus.FAILED;
        }

        SSHClient ssh = new SSHClient();
        ssh.setTimeout(SshProxySocketImplFactory.KEX_TIMEOUT_MS);
        ssh.setConnectTimeout(SshProxySocketImplFactory.CONNECT_TIMEOUT_MS);
        ssh.addHostKeyVerifier(new PromiscuousVerifier());

        try {
            SocketFactory socketFactory = proxySocketFactoryProvider.createProxySocketFactory(loginService.getHostname(), loginService.getPort());
            if (socketFactory != null) {
                LOG.debug("Using proxy for SSH access");
                ssh.setSocketFactory(socketFactory);
            } else {
                LOG.debug("No proxy needed for SSH access");
            }
        } catch (IOException ex) {
            LOG.error("Error while creating ProxySocketFactory. Trying without now.", ex);
        }

        int connectTries = 0;
        boolean connected = false;
        Exception lastConnectEx = null;

        while(!connected && connectTries < SETUP_SSH_CONNECTION_RETRIES){
            connectTries++;
            try {
                ssh.connect(loginService.getHostname(), loginService.getPort());
                connected = true;
            } catch (IOException e) {
                lastConnectEx = e;
                String message = "Error while connecting to " + loginService.getHostname() + ": " + e.getMessage();
                if(connectTries < SETUP_SSH_CONNECTION_RETRIES)
                    message += ". Retrying ...";
                updateMessage(message);
                LOG.error("Error while connecting to {}: {}", loginService.getHostname(), e.getMessage(), e);
            }
        }

        if(!connected){
//            return ExperimentTaskStatus.FAILED;
            throw new SshSetupException("SSHClient could not connect after "+connectTries+" tries.", lastConnectEx);
        }

        try {
            //we must use the distributed ssh keys to login: in speaksfor case, there is no other way to login
            ssh.authPublickey(loginService.getUsername(), new KeyPairWrapper(loginKeyPair));
        } catch (UserAuthException e) {
            updateMessage("Could not authenticate: " + e.getMessage());
            LOG.error("Authentication failed on {}: {}", node.getClientId(), e.getMessage(), e);
//            return ExperimentTaskStatus.FAILED;
            throw new SshSetupException("Authentication failed on "+node.getClientId(), e);
        } catch (TransportException e) {
            updateMessage("Error while authenticating: " + e.getMessage());
            LOG.error("Error while authenticating on {}: {}", node.getClientId(), e.getMessage(), e);
//            return ExperimentTaskStatus.FAILED;
            throw new SshSetupException("Error while authenticating on "+node.getClientId(), e);
        }

        return ssh;
    }


    private Collection<SfaExperimentPart> getSfaParts() {
        return experiment.getPartsStream()
                .filter(part -> part instanceof SfaExperimentPart)
                .map(SfaExperimentPart.class::cast)
                .collect(Collectors.toList());

    }

    private Collection<NonSfaExperimentPartController> getNonSfaExperimentPartControllers() {

        Set<NonSfaExperimentPartController> controllers = new HashSet<>();

        for (ExperimentPart experimentPart : experiment.getPartsListCopy()) {
            if (experimentPart instanceof SfaExperimentPart)
                continue;


            ExperimentPartController epc = epcManager.getController(experimentPart);
            if (!(epc instanceof NonSfaExperimentPartController))
                throw new RuntimeException("Expected a non-sfa ExperimentPartController");

            controllers.add((NonSfaExperimentPartController) epc);
        }

        return controllers;
    }

    private boolean hasNonSfaParts() {
        return experiment.getPartsStream().anyMatch(part -> !(part instanceof SfaExperimentPart));
    }

    private boolean hasSfaParts() {
        return experiment.getPartsStream().anyMatch(part -> (part instanceof SfaExperimentPart));
    }
}
